﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;
using Buddy.Coroutines;
using ff14bot;
using ff14bot.Behavior;
using ff14bot.Helpers;
using ff14bot.Managers;
using ff14bot.Navigation;
using TreeSharp;

namespace FateStayTeleport
{
    internal enum Mode
    {
        Leveling,
        AtmaHunting,
        Luminous
    }

    internal class TeleportLogic
    {
        private Composite behavior;
        private readonly List<ZoneData> zones;
        private readonly Mode mode;
        private ZoneData targetZone;
        private const uint TICK_TIME = 5000;
        private readonly Stopwatch tickWatch = new Stopwatch();
        private readonly HashSet<uint> atmaIds;
        private readonly int atmaAmount;
        private readonly int luminousAmount;
        private bool startWithCheck = true;
        
        public TeleportLogic(TeleportSettings settings)
        {
            atmaAmount = settings.AtmaAmount;
            luminousAmount = settings.LuminousAmount;
            zones = settings.Zones.ToList();
            mode = settings.Mode;

            if (mode == Mode.AtmaHunting)
            {
                var atmaArray = zones
                    .Where(z => z.Atma != 0)
                    .Select(z => z.Atma)
                    .ToArray();
                atmaIds = new HashSet<uint>(atmaArray);
            }
            else
            {
                var atmaArray = zones
                    .Where(z => z.Luminous != 0)
                    .Select(z => z.Luminous)
                    .ToArray();
                atmaIds = new HashSet<uint>(atmaArray);
            }

            
        }

        public Composite GetBehavior()
        {
            if (behavior != null) { return behavior; }

            var coroutine = new ActionRunCoroutine(r => CheckTeleport());
            behavior = new Decorator(c => Tick(), coroutine);

            return behavior;
        }

        private bool Tick()
        {
            if (!tickWatch.IsRunning) { tickWatch.Start(); }
            if (tickWatch.ElapsedMilliseconds < TICK_TIME && !startWithCheck) { return false; }

            tickWatch.Reset();
            startWithCheck = false;

            return true;
        }

        private async Task<bool> CheckTeleport()
        {
            if (!CanTeleport || !TeleportNeeded()) { return false; }
            return await Teleport();
        }

        private async Task<bool> Teleport()
        {
            uint zone = targetZone.Zone;
            uint aetheryte = targetZone.Aetheryte;

            Logger.Log("Teleporting to " + targetZone.Name + ".");
            Poi.Clear("We're teleporting to another zone.");
            Navigator.Stop();

            while (CanTeleport && WorldManager.ZoneId != zone)
            {
                if (Core.Me.IsMounted) { await CommonTasks.StopAndDismount(); }
                if (!Core.Me.IsCasting && !CommonBehaviors.IsLoading) { WorldManager.TeleportById(aetheryte); }
                await CarefulSleep(500);
            }

            return true;
        }

        private static async Task<bool> CarefulSleep(uint ms)
        {
            var stopwatch = new Stopwatch();
            stopwatch.Start();

            while (stopwatch.ElapsedMilliseconds < ms && CanTeleport)
            {
                await Coroutine.Yield();
            }
            stopwatch.Stop();

            return CanTeleport;
        }

        private static bool CanTeleport
        {
            get
            {
                return !Core.Me.IsDead
                       && !Core.Me.InCombat
                       && !FateManager.WithinFate;
            }
        }

        private bool TeleportNeeded()
        {
            if (mode == Mode.Leveling)
                targetZone = TargetByLeveling();
            else if (mode == Mode.AtmaHunting)
                targetZone = TargetByAtma();
            else
                targetZone = TargetByLuminous();

            return targetZone != null && targetZone != GetCurrentZone();
        }

        private ZoneData TargetByLeveling()
        {
            uint level = Core.Me.ClassLevel;

            return zones
                .Where(z => level <= z.MaxLevel && level >= z.MinLevel && !z.Ignore)
                .FirstOrDefault(z => WorldManager.HasAetheryteId(z.Aetheryte));
        }

        private ZoneData TargetByAtma()
        {

            var finishedAtmas = InventoryManager.FilledSlots
                .Where(s => atmaIds.Contains(s.RawItemId))
                .Where(s => s.Count >= atmaAmount)
                .Select(s => s.RawItemId);

            var finishedSet = new HashSet<uint>(finishedAtmas);
            var date = DateTime.UtcNow.AddHours(9);
            int time = Convert.ToInt32(date.ToString("%h"));

            var hourZone = zones
                .Where(z => z.Atma != 0)
                .Where(z => WorldManager.HasAetheryteId(z.Aetheryte) && !z.Ignore)
                .Where(z => !finishedSet.Contains(z.Atma))
                .OrderByDescending(z => z.AtmaHour == time)
                .FirstOrDefault();

            return hourZone;
        }

        private ZoneData TargetByLuminous()
        {

            var finishedLuminous = InventoryManager.FilledSlots
                .Where(s => atmaIds.Contains(s.RawItemId))
                .Where(s => s.Count >= luminousAmount)
                .Select(s => s.RawItemId);

            var finishedSet = new HashSet<uint>(finishedLuminous);

            return zones
                .Where(z => z.Luminous != 0)
                .Where(z => WorldManager.HasAetheryteId(z.Aetheryte) && !z.Ignore)
                .Where(z => !finishedSet.Contains(z.Atma))
                .FirstOrDefault(z => WorldManager.HasAetheryteId (z.Aetheryte));
        }


        private ZoneData GetCurrentZone()
        {
            return zones
                .FirstOrDefault(z => z.Zone == WorldManager.ZoneId);
        }
    }
}
